# -*- coding: utf-8 -*- 

import xapian
from utils import clean_value, clean_date

_qp_flags_base = xapian.QueryParser.FLAG_LOVEHATE
_qp_flags_phrase = xapian.QueryParser.FLAG_PHRASE
_qp_flags_synonym = (xapian.QueryParser.FLAG_AUTO_SYNONYMS |
                     xapian.QueryParser.FLAG_AUTO_MULTIWORD_SYNONYMS)
_qp_flags_bool = xapian.QueryParser.FLAG_BOOLEAN

class Query(object):

    """ query class """

    def __init__(self, schema):
        """ 初始化 """
        self.schema = schema
        self._filters = []
        self._exclude = []
        self._combine_operation = 'and'
        self._combine_query = None

    def _combine(self, other, operation):
        if not isinstance(other, type(self)):
            raise TypeError, "the other query must be Query instance"

        if self._combine_query is not None:
            self._combine_query._combine(other, operation)
        else:
            self._combine_operation = operation
            self._combine_query = other
            return self

    def __or__(self, other):
        """
           逻辑或重载，用来将两个QuerySet 用OR合并
           **第二个QuerySet 的某些参数将被忽略(start, limit, restricted, sort, unique)**
           RETURN一个新的QuerySet 对象
        """
        return self._combine(other, 'or')

    def __and__(self, other):
        """
           逻辑或重载，用来将两个QuerySet 用and合并
           **第二个QuerySet 的某些参数将被忽略(start, limit, restricted, sort, unique)**
           RETURN一个新的QuerySet 对象
        """
        return self._combine(other, 'and')

    def exclude(self, key, value, op):
        """ 排除条件符合条件的结果 """
        return self.filter(key, value, op, exclude=True)

    def parse(self, text, fields):
        """ 跨字段搜索函数, 可以使用通配符
            - fields 允许搜索的字段，为None则搜索全部字段
        """
        # 搜索支持部分匹配
        self._filters.append((fields, text, 'parse'))
        return self

    def filter(self, key, value, op, exclude=False):
        """ 添加条件

            - ops
            支持的操作符：

            - range   进行range查询(只有数字、日期字段才可进行range查询)
            - allof
            - anyof
        """
        if exclude:
            self._exclude.append((key, value, op))
        else:
            self._filters.append((key, value, op))

        return self

    def _get_xapian_query(self, querys, exclude=False, database=None):
        """ convert to xapian query 
        combined = query_filter(self.query_restricted(), combined)
        """
        if not querys:
            return xapian.Query('')

        qp = xapian.QueryParser()
        if database is not None:
            qp.set_database(database)
        qp.set_default_op(xapian.Query.OP_AND)

        # parse filters
        queries = []
        for filters in querys:
            field, value, op = filters
            if op == 'parse':
                _queries = []
                value = clean_value(value, is_query=True)
                for f in field:
                    prefix = self.schema.get_prefix(f, auto_add=False)
                    _queries.append( qp.parse_query(value, xapian.QueryParser.FLAG_WILDCARD, prefix) )

                query = xapian.Query(xapian.Query.OP_OR, _queries)

                queries.append(query)
                continue

            if not value:
                continue

            if op == 'allof':
                prefix = self.schema.get_prefix(field, auto_add=False)
                value = clean_value(value)
                query = query_field(prefix, value)
                queries.append(query)

            elif op == 'anyof':
                prefix = self.schema.get_prefix(field, auto_add=False)
                value = clean_value(value)
                query = query_field(prefix, value, default_op=xapian.Query.OP_OR)
                queries.append(query)

            elif op == 'range':
                prefix = self.schema.get_slot(field, auto_add=False)
                value = clean_date(value)
                begin, end = value[:2]
                query = self.query_range(prefix, begin, end)
                queries.append(query)

            elif not op:
                prefix = self.schema.get_prefix(field, auto_add=False)
                value = clean_value(value)
                query = query_field(prefix, value)
                queries.append(query)

        if len(queries) == 1:
            combined = queries[0]
        else:
            if not exclude:
                _func = lambda q1, q2: query_filter(q1, q2)
                combined = reduce( _func, queries)
            else:
                combined = xapian.Query(xapian.Query.OP_OR, list(queries))

        return combined

    def query_range(self, slot, begin, end):
        """ """
        if begin is None and end is None:
            # Return a "match everything" query
            return xapian.Query('')

        if slot is None:
            # Return a "match nothing" query
            return xapian.Query()

        begin, end = normalize_range(begin, end)

        if begin is None:
            return xapian.Query(xapian.Query.OP_VALUE_LE, slot, end)

        if end is None:
            return xapian.Query(xapian.Query.OP_VALUE_GE, slot, begin)

        return xapian.Query(xapian.Query.OP_VALUE_RANGE, slot, begin, end)

    def unique(self, field):
        """ 对指定字段，如有多个相同的值则返回结果中只出现一次

            **只有设置了collapse 的字段才可进行unique查询**
        """
        self._collapse = field
        return self

    def to_dict(self):
        """  将query对象转换为基本的python字典对象
        """
        # TODO

    def build_query(self, database=None):
        """ 生成xappy查询
            RETURN Xapian Query 对象

            database: xapian 使用通配符模式搜索，需要指定数据库对象
                      如果你使用了parse方法，需要在这里指定数据库对象
        """
        source = self
        other = self._combine_query
        operation = self._combine_operation

        # 先对自己build query
        combined = source._get_xapian_query(source._filters, database=database)
        # 添加exclude的查询条件
        if source._exclude:
            exclude_query = source._get_xapian_query(source._exclude, True, database=database)
            if exclude_query:
                combined = query_filter(combined, exclude_query, exclude=True)

        # 递归合并多个Query, 只允许and, or两种操作
        while other is not None:
            combined2 = other._get_xapian_query(other._filters, database=database)
            # 添加exclude的查询条件
            if self._exclude:
                exclude_query = other._get_xapian_query(other._exclude, True, database=database)
                if exclude_query:
                    combined2 = query_filter(combined2, exclude_query, exclude=True)

            if operation == 'and':
                combined = query_filter(combined, combined2)
            else:
                combined = xapian.Query(xapian.Query.OP_OR, [combined, combined2])

            operation = other._combine_operation
            other = other._combine_query

        return combined

def normalize_range(begin, end):
    """ 查询时，转换range 参数，主要是把 float/int 转换为 str 格式 """

    if begin is not None:
        if isinstance(begin, float):
            begin = xapian.sortable_serialise(float(begin))
        else:
            begin = str(begin)

    if end is not None:
        if isinstance(end, float):
            end = xapian.sortable_serialise(float(end))
        else:
            end = str(end)
    return begin, end

def _query_parse_with_prefix(qp, string, flags, prefix):
    """ """
    if prefix is None:
        return qp.parse_query(string, flags)
    else:
        return qp.parse_query(string, flags, prefix)

def query_filter(query, filter, exclude=False):
    """ """
    if not isinstance(filter, xapian.Query):
        raise Exception("filter must be a xapian query object")
    if exclude:
        return xapian.Query(xapian.Query.OP_AND_NOT, query, filter)
    else:
        return xapian.Query(xapian.Query.OP_FILTER, query, filter)

def query_field(prefix, value, default_op=xapian.Query.OP_AND):
    """ """ 
    #if types == 'exact':
    #    prefix = catalog.fields[field]['prefix']
    #    if len(value) > 0:
    #        chval = ord(value[0])
    #        if chval >= ord('A') and chval <= ord('Z'):
    #            prefix = prefix + ':'
    #    return xapian.Query(prefix + value)

    #if types == 'freetext':
    #    qp = xapian.QueryParser()
    #    qp.set_default_op(default_op)
    #    prefix = self.get_prefix(field)
    #    return _query_parse_with_fallback(qp, value, prefix)

    #return xapian.Query()


    if not prefix:
        return xapian.Query()
    else:
        qp = xapian.QueryParser()
        qp.set_default_op(default_op)
        return _query_parse_with_fallback(qp, value, prefix)

def _query_parse_with_fallback(qp, string, prefix=None):
    """ """
    try:
        q1 = _query_parse_with_prefix(qp, string,
                                       _qp_flags_base |
                                       _qp_flags_phrase |
                                       _qp_flags_synonym |
                                       _qp_flags_bool,
                                       prefix)
    except xapian.QueryParserError:
        # If we got a parse error, retry without boolean operators (since
        # these are the usual cause of the parse error).
        q1 = _query_parse_with_prefix(qp, string,
                                           _qp_flags_base |
                                           _qp_flags_phrase |
                                           _qp_flags_synonym,
                                           prefix)

    qp.set_stemming_strategy(qp.STEM_NONE)
    try:
        q2 = _query_parse_with_prefix(qp, string,
                                           _qp_flags_base |
                                           _qp_flags_bool,
                                           prefix)
    except xapian.QueryParserError:
        # If we got a parse error, retry without boolean operators (since
        # these are the usual cause of the parse error).
        q2 = _query_parse_with_prefix(qp, string,
                                           _qp_flags_base,
                                           prefix)

    return xapian.Query(xapian.Query.OP_AND_MAYBE, q1, q2)




# -*- encoding: utf-8 -*-
""" zapian api 
"""

import os
import gc
import shutil
import xapian
import time
import hashlib
import cPickle as pickle
from threading import local
from datetime import datetime

from query import Query
from utils import clean_value
from schema import Schema, CONFIG_FILE

class Zapian(object):

    def __init__(self, db_path):
        self.db_path = db_path
        self.parts = [] 
        for part_name in os.listdir(self.db_path):
            self.add_part(part_name)

        self.schema = Schema(db_path)

    def add_part(self, part_name):
        """ add xapian a database """
        if part_name in self.parts:
            return 

        part_path = os.path.join(self.db_path, part_name)
        if os.path.isdir(part_path):
            self.parts.append(part_name)

    def release_part(self, part_name):
        """ release xapian a database """
        _release_write_db(self.db_path, part_name)

    def remove_part(self, part_name):
        """ remove xapian a database """
        part_path = os.path.join(self.db_path, part_name)
        if os.path.isdir(part_path):
            part = _write_database_index.pop(part_path, None)
            part.close()
            shutil.rmtree(part_path)
        else:
            raise Exception('remove database: %s is not xapian database'%part_path)

    def get_interior_doc(self, doc, data=None, old_doc=None):
        """ convert python dict into xapian document object
            doc: 
               {field1:value1, 
                field2:value2}

            data: raw data for original object

            return: xapian document object
        """
        def _add_term(doc, termgen, prefix, value):
            type_name = 'freetext'

            if type_name == 'exact':
                if len(value) > 0:
                    # We use the following check, rather than "isupper()" to ensure
                    # that we match the check performed by the queryparser, regardless
                    # of our locale.
                    if ord(value[0]) >= ord('A') and ord(value[0]) <= ord('Z'):
                        prefix = prefix + ':'

                # Note - xapian currently restricts term lengths to about 248
                # characters - except that zero bytes are encoded in two bytes, so
                # in practice a term of length 125 characters could be too long.
                # Xapian will give an error when commit() is called after such
                # documents have been added to the database.
                # As a simple workaround, we give an error here for terms over 220
                # characters, which will catch most occurrences of the error early.
                #
                # In future, it might be good to change to a hashing scheme in this
                # situation (or for terms over, say, 64 characters), where the
                # characters after position 64 are hashed (we obviously need to do this
                # hashing at search time, too).
                if len(prefix + value) > 220:
                    raise Exception("Field is too long: maximum length "
                                               "220 - was %d (%r)" %
                                               (len(prefix + value),
                                                prefix + value))


                doc.add_term(prefix + value, 1) # wdfinc default set 1

            elif type_name == 'freetext':
                # no positions, weight default set 1
                termgen.index_text_without_positions(str(value), 1, prefix)
                termgen.increase_termpos(10)

        def _add_value(doc, slotnum, value):

            if isinstance(value, float):
                value = xapian.sortable_serialise(float(value))
                doc.add_value(int(slotnum), value)
            else:
                doc.add_value(int(slotnum), str(value))

        document = xapian.Document()
        termgen = xapian.TermGenerator()
        termgen.set_document(document)

        terms = set()
        values = set()
        # build new xapian object
        for field, value in doc.iteritems():
            both = field.startswith('+') 
            if both:
                field = field[1:]
            is_value = isinstance(value, (int, float, datetime))
            is_term = not is_value
            # sortable
            if both or is_value:
                if field in values:
                    continue
                slotnum = self.schema.get_slot(field)
                value = clean_value(value)
                _add_value(document, slotnum, value)
                values.add(slotnum)
            # field
            if both or is_term:
                if field in terms:
                    continue
                prefix = self.schema.get_prefix(field)
                value = clean_value(value)
                _add_term(document, termgen, prefix, value)
                terms.add(prefix)

        # new value will be replace old value
        if old_doc is not None:
            for term in old_doc.termlist():
                prefix, value = self.schema.split_term(term.term)
                if prefix not in terms:
                    _add_term(document, termgen, prefix, value)

            for value in old_doc.values():
                if value.num not in values:
                    _add_value(document, value.num, value.value)

            if data is None: 
                data = dict()

            old_data = old_doc.get_data()
            if old_data:
                old_data = pickle.loads(old_data)
                for k, v in old_data.iteritems():
                    if k not in data:
                        data[k] = v
        # add data
        if data:
            document.set_data(pickle.dumps(data))

        return document

    def add_document(self, part_name, index, uid=None, data=None, flush=True):
        """ add a xapian document
        """
        db = _get_write_db(self.db_path, part_name)
        doc = self.get_interior_doc(index, data=data)
        if not uid:
            identifier = u'Q' + part_name + str(db.get_lastdocid())
        else:
            identifier = u'Q' + str(uid)
        doc.add_boolean_term(identifier)
        db.replace_document(identifier, doc)
        if flush: db.commit()

    def replace_document(self, part_name, uid, index, data=None, flush=True):
        """ replace existing xapian document """
        doc = self.get_interior_doc(index, data=data)
        db = _get_write_db(self.db_path, part_name)
        identifier = u'Q' + str(uid)
        doc.add_boolean_term(identifier)
        db.replace_document(identifier, doc)
        if flush: db.commit()

    def delete_document(self, part_name, uids, flush=True):
        """ delete a xapian document """
        if not isinstance(uids, (list, tuple, set)):
            uids = (uids,)
        db = _get_write_db(self.db_path, part_name)
        for uid in uids:
            identifier = u'Q' + str(uid)
            db.delete_document(identifier)
        if flush: db.commit()

    def update_document(self, part_name, uid, index, data=None, flush=True):
        """ update xapian document existing fields and attributes
        """
        db = _get_write_db(self.db_path, part_name)

        identifier = u'Q' + str(uid)
        old_doc = self._get_document(str(uid), [part_name], db=db)

        new_doc = self.get_interior_doc(index, data=data, old_doc=old_doc)
        
        db.replace_document(identifier, new_doc)
        if flush: db.commit()

    def get_document(self, uid, part=None):
        """ Get the document """
        xap_doc = self._get_document(uid, part)
        return pickle.loads( xap_doc.get_data() )

    def _get_document(self, uid, part=None, db=None):
        """Get the xapian document object with the specified unique ID.

        Raises a KeyError if there is no such document.  Otherwise, it returns
        a ProcessedDocument.

        """
        if db is None:
            db = _get_read_db(self.db_path, part or self.parts)

        while True:
            try:
                postlist = db.postlist('Q' + uid)
                try:
                    plitem = postlist.next()
                except StopIteration:
                    # Unique ID not found
                    raise KeyError('Unique ID %r not found' % uid)
                try:
                    postlist.next()
                    raise Exception("Multiple documents " #pragma: no cover
                                               "found with same unique ID")
                except StopIteration:
                    # Only one instance of the unique ID found, as it should be.
                    pass

                return db.get_document(plitem.docid)

            except xapian.DatabaseModifiedError:
                reopen(db)

    def commit(self, part_name=None):
        """ commit xapian database """
        if part_name is None:
            for part_name in self.parts:
                db = _get_write_db(self.db_path, part_name)
                db.commit()
        else:
            db = _get_write_db(self.db_path, part_name)
            db.commit()

    def search(self, parts=None, query=None, query_obj=None, orderby=None, start=None, stop=None):
        """ 搜索, 返回document id的集合 

        如果parts为空，会对此catalog的所有索引进行搜索。

        """
        # 这个目录不一个正确的数据库，可能还没有保存至少一条数据
        if CONFIG_FILE not in os.listdir(self.db_path):
            return []

        if parts is None:
            parts = self.parts

        else:
            for part_name in parts:
                if not os.path.exists(os.path.join(self.db_path, part_name)):
                    parts.remove(part_name)

        if not parts:
            return []

        database = _get_read_db(self.db_path, parts=parts)

        if query_obj is not None:
            xapian_query = query_obj.build_query(database=database)
        else:
            query_obj = Query(self.schema)
            query_obj._filters = query
            xapian_query = query_obj.build_query(database=database)

        enquire = xapian.Enquire(database)
        enquire.set_query(xapian_query)
        
        # sort
        if orderby is not None:
            asc = True
            if orderby[0] == '-':
                asc = False
                orderby = orderby[1:]
            elif orderby[0] == '+':
                orderby = orderby[1:]

            slotnum = self.schema.get_slot(orderby, auto_add=False)
            # sort when available
            if slotnum is not None:
                # Note: we invert the "asc" parameter, because xapian treats
                # "ascending" as meaning "higher values are better"; in other
                # words, it considers "ascending" to mean return results in
                # descending order.
                enquire.set_sort_by_value_then_relevance(slotnum, not asc)

        enquire.set_docid_order(enquire.DONT_CARE)

        # Repeat the search until we don't get a DatabaseModifiedError
        if start is None: start = 0
        if stop is None: stop = database.get_doccount()
        while True:
            try:
                matches = enquire.get_mset(start, stop)
                break
            except xapian.DatabaseModifiedError:
                reopen(database)

        # 返回结果的ID集
        def _get_docid(match):
            tl = match.document.termlist()
            try:
                term = tl.skip_to('Q').term
                if len(term) == 0 or term[0] != 'Q':
                    return None
            except StopIteration:
                return None
            return term[1:]

        return map(_get_docid, matches)

_write_database_index = {}
def _get_write_db(db_path, part_name, protocol=''):
    """ get xapian writable database 
        protocol: the future may support.
    """
    part_path = os.path.join(db_path, part_name)
    # writeable database is already open, this will raise a xapian.DatabaseLockError
    # so, writeable database need to cached.
    if part_path in _write_database_index:
        return _write_database_index[part_path]
    else:
        db = xapian.WritableDatabase(part_path, xapian.DB_CREATE_OR_OPEN)
        _write_database_index[part_path] = db
        return db

def _release_write_db(db_path, part_name, protocol=''):
    """ """
    part_path = os.path.join(db_path, part_name)
    if part_path in _write_database_index:
        _write_database_index[part_path].close()
        del _write_database_index[part_path]

thread_context = local()
READ_DB_REFRESH_DELTA = 2 # max time in seconds till we refresh a connection
def _get_read_db(db_path, parts, protocol=''):
    """ get xapian readonly database
        protocol: the future maybe support.
    """
    if not getattr(thread_context, 'connection', None):
        thread_context.connection = {}

    if not getattr(thread_context, 'modified', None):
        thread_context.modified = {}

    if not getattr(thread_context, 'opened', None):
        thread_context.opened = {}

    prefix = hashlib.md5(db_path + ''.join(parts)).hexdigest()
    conn = thread_context.connection.get(prefix, None)
    now = time.time()

    if thread_context.modified.get(prefix, 0) + READ_DB_REFRESH_DELTA < now:
        thread_context.modified[prefix] = now
    
    if conn is None:
        base_name = os.path.join(db_path, parts[0])
        conn = xapian.Database(base_name)

        # 适用于多个数据库查询
        for part_name in parts[1:]:
            other_name = os.path.join(db_path, part_name)
            conn.add_database(xapian.Database(other_name))
        thread_context.connection[prefix] = conn 
        thread_context.opened[prefix] = now
            
    opened = thread_context.opened.get(prefix, 0)

    if opened < thread_context.modified[prefix]:
        reopen(conn)
        thread_context.opened[prefix] = now

    return conn

def _release_read_db(db_path, parts, protocol=''):
    """ release the read db 
        protocol: the future maybe support.
    """
    prefix = hashlib.md5(db_path + ''.join(parts)).hexdigest()
    del thread_context.modified[prefix]
    del thread_context.opened[prefix]
    thread_context.connection[prefix].close()
    del thread_context.connection[prefix]

def reopen(db):
    db.reopen()
    gc.collect()



# -*- encoding: utf-8 -*-
"""
CJKSplitter - Chinese, Japanese, Korean word splitter for ZCTextIndex

(C) by www.zopen.cn, panjy@zopen.cn & others

License: see LICENSE.txt

$Id: Exp $

DESCRIPTION OF ALGORITHM

  Continuous letters and numbers make up words.  Spaces and symbols
  separate letters and numbers into words.  This is sufficient for
  all western text.

  CJK doesn't use spaces or separators to separate words, so the only
  way to really find out what constitutes a word would be to have a
  dictionary and advanced heuristics.  Instead, we form pairs from
  consecutive characters, in such a way that searches will find only
  characters that appear more-or-less the right sequence.  For example:

    ABCDE => AB BC CD DE

  This works okay since both the index and the search query is split
  in the same manner, and since the set of characters is huge so the
  extra matches are not significant.
"""
#try:
#    from Products.ZCTextIndex.ISplitter import ISplitter
#    from Products.ZCTextIndex.PipelineFactory import element_factory
#except:
#    pass
#from types import StringType

import re

def getSupportedEncoding(encodings):
    for encoding in encodings:
        try:
            unicode('A', encoding)
            return encoding
        except:
            pass
    return 'utf-8'

# CJK charsets ranges, see this following pages:
#
# http://jrgraphix.net/research/unicode_blocks.php?block=87
# http://jrgraphix.net/research/unicode_blocks.php?block=85
# http://jrgraphix.net/research/unicode_blocks.php?block=95
# http://jrgraphix.net/research/unicode_blocks.php?block=76
# http://jrgraphix.net/research/unicode_blocks.php?block=90

# more to support:
# \u30a1-\u30f6 : Japanese Katakana
# \uff10-\uff19 : full-width number
# \uff21-\uff3a : full-width alphabete
# \uff41-\uff5a
# (done) \u0392-\u03c9 : greek.

norm_table = {'　':' '}

for char_ord in range(ord(u'Ａ'), ord(u'Ｚ')+1):
    norm_table[unichr(char_ord)] = unichr(ord('A') + char_ord - ord(u'Ａ'))
for char_ord in range(ord(u'ａ'), ord(u'ｚ')+1):
    norm_table[unichr(char_ord)] = unichr(ord('a') + char_ord - ord(u'ａ'))
for char_ord in range(ord(u'０'), ord(u'９')+1):
    norm_table[unichr(char_ord)] = unichr(ord('0') + char_ord - ord(u'０'))

rxNormal = re.compile(u"[a-zA-Z0-9_\u0392-\u03c9]+|[\u4E00-\u9FFF\u3400-\u4dbf\uf900-\ufaff\u3040-\u309f\uac00-\ud7af]+", re.UNICODE)
rxGlob = re.compile(u"[a-zA-Z0-9_\u0392-\u03c9]+[*?]*|[\u4E00-\u9FFF\u3400-\u4dbf\uf900-\ufaff\u3040-\u309f\uac00-\ud7af]+[*?]*", re.UNICODE)

class CJKSplitter(object):

    def process(self, lst, isGlob=0, deep_split_english=False):
        result = []
        if isGlob:
          rx = rxGlob
        else:
          rx = rxNormal
        for s in lst:
            # s must be unicode!

            # normalize
            new_s = []
            for char in s:
                if char in norm_table:
                    new_s.append(norm_table[char])
                else:
                    new_s.append(char)
            s = ''.join(new_s)

            splitted = rx.findall(s)
            for w in splitted:


                if ord(w[0]) >= 12352:  # \u3040
                    len_w = len(w)
                    if len_w == 1:
                        if isGlob:
                            result.append(w + '*')
                        else:
                            result.append(w )
                    elif len_w == 2:
                        result.append(w)
                        # XXX here, we may lost some index, due to ZCTextIndex's phrase process
                        # ZCTextIndex will check each processGlob 'ed words again. And it doesn't
                        # work with CJKSplitter. It will break the search term
                        # so I have to comment the 2 lines.
                        # if not isGlob:
                        #     result.append(w[1])
                    else:
                        i = 2
                        while i <= len_w:
                            result.append(w[i-2:i])
                            # result.append(w[i-1:i+1])
                            i += 1

                        if not isGlob:
                            result.append(w[-1])
                else:
                    if deep_split_english:
                        # title 进索引的时候，只要把英文和数字分词就行
                        result.extend(self.__process_one_word(w))
                    else:
                        # title 搜索关键字，不需要分词，但是要在最后加上*，以便支持模糊搜索
                        result.append(w if not isGlob else w + '*')
        # return [word.encode('utf8') for word in result]
        return result

    def processGlob(self, lst):
        return self.process(lst, 1)

    def __process_one_word(self, word):
        """ 英文数字分词 """
        word_list = list(word)
        word_len = len(word_list)

        ### 只有一位或者两位
        if word_len <= 2:
            return [word]

        ### 大于两位

        results = []
        result = word_list
        results.append(''.join(word_list))

        # abcd
        # bcd
        # bc
        # 每次让前面pop出一位, 直到只剩下两位
        while 1:
            result.pop(0)
            results.append(''.join(result))
            if len(result) <= 2: break

        return results

if __name__ == '__main__':
   words = ['我们abs非常好ddd ａｂｃ　ｄｅｆ',
        u'"我们非常好 1212* 有033212-1" and niubi' ,]

   for word in words:
       print '=====now test:', word
       s = CJKSplitter()
       print 'no glob result:'
       for i in s.process([word], deep_split_english=True):
           print i

       print 'glob result:'
       for i in s.processGlob([word]):
           print i




#encoding=utf-8

import os
import shutil
import cPickle as pickle
import tempfile
import unittest
from datetime import datetime

from api import Zapian
from query import Query

# not cached only read database connection of xapian
import api
api.READ_DB_REFRESH_DELTA = 0

class ZapianTest(unittest.TestCase):

    def setUp(self):
        # init the database attributes
        self.data_root = tempfile.mkdtemp()
        self.parts = ['part01', 'part02']

        # init the database environ
        if not os.path.exists(self.data_root):
            print 'create %s database' % self.data_root
            os.makedirs(self.data_root)

        self.zapian = Zapian(self.data_root)
        for part_name in self.parts:
            database_path = os.path.join(self.data_root, part_name)
            if not os.path.isdir(database_path):
                os.makedirs(database_path)
            self.zapian.add_part(part_name)

        # init the test data
        self.doc = {'+title':'we are firend', 
                    'subjects':['file','ktv','what@gmail.com'], 
                    'created':datetime(2000, 1, 1)}

    def tearDown(self):
        """ """
        if os.path.exists(self.data_root):
            shutil.rmtree(self.data_root)

        self.zapian.schema.fields.clear()
        self.zapian.schema.attributes.clear()

    def test_schema(self):
        schema = self.zapian.schema
        # test gen prefix and slot
        prefix = schema._gen_prefix() 
        self.assertEqual(prefix, 'XA')
        slot = schema._gen_slot()
        self.assertEqual(slot, 0)
        # test add field and attribute
        new_prefix = schema.add_field('new_field')
        self.assertTrue(schema.get_prefix('new_field') == new_prefix == prefix)
        new_slot = schema.add_attribute('new_attribute')
        self.assertTrue(schema.get_slot('new_attribute') == new_slot == slot)
        # test load the old schema
        new_zapian = Zapian(self.data_root)
        self.assertEqual(new_zapian.schema.fields, {'new_field':'XA'})
        self.assertEqual(new_zapian.schema.attributes, {'new_attribute':0})

    def test_add_document(self):
        part = self.parts[0]
        uid = "12345"
        self.zapian.add_document(part, uid=uid, 
                index=self.doc, data={'data': "测试内容"},
                flush=True)
        # test value of the new document
        doc = self.zapian._get_document(uid, [part])
        values = [v.value for v in doc.values()]
        self.assertEqual(set(values), set(['946656000', 'we are firend']))
        # test term of the new document
        title_prefix = self.zapian.schema.get_prefix('title', auto_add=False)
        subjects_prefix = self.zapian.schema.get_prefix('subjects', auto_add=False)
        validate_terms = ['Q'+uid,
                            title_prefix + 'are', 
                            title_prefix + 'firend', 
                            title_prefix + 'we',
                            subjects_prefix + 'file', 
                            subjects_prefix + 'ktv', 
                            subjects_prefix + 'what_gmail_com', ]
        old_terms = [ term.term for term in doc.termlist() ]
        self.assertEqual(len(validate_terms), len(old_terms))
        self.assertEqual(set(validate_terms), set(old_terms))
        # test data of the new document
        data = pickle.loads( doc.get_data() )['data']
        self.assertEqual(data, '测试内容')

    def test_update_document(self):
        part = self.parts[0]
        uid = "12345"
        # add a document 
        self.zapian.add_document(part, uid=uid, 
                index=self.doc, data={'data': "测试内容"},
                flush=True)
        new_doc = self.doc.copy()
        new_doc['+title'] = "new title"

        self.zapian.update_document(part, uid=uid, index=new_doc, flush=True)
        # test value of the new document
        doc = self.zapian._get_document(uid, [part])
        values = [v.value for v in doc.values()]
        self.assertEqual(set(values), set(['946656000', 'new title']))
        # test term of the new document
        title_prefix = self.zapian.schema.get_prefix('title', auto_add=False)
        subjects_prefix = self.zapian.schema.get_prefix('subjects', auto_add=False)
        validate_terms = ['Q'+uid,
                            title_prefix + 'new', 
                            title_prefix + 'title', 
                            subjects_prefix + 'file', 
                            subjects_prefix + 'ktv', 
                            subjects_prefix + 'what_gmail_com', ]
        old_terms = [term.term for term in doc.termlist()]
        self.assertTrue(len(validate_terms) == len(old_terms))
        self.assertEqual(set(validate_terms), set(old_terms))
        # test data of the new document
        data = pickle.loads( doc.get_data() )['data']
        self.assertEqual(data, '测试内容')

    def test_replace_document(self):
        part = self.parts[0]
        uid = "12345"
        # add a document 
        self.zapian.add_document(part, uid=uid, 
                index=self.doc, data={'data': "测试内容"},
                flush=True)
        new_doc = self.doc.copy()
        new_doc['+title'] = "new title"
        self.zapian.schema.add_field('new-field')
        new_doc['new-field'] = 'last'

        self.zapian.replace_document(part, uid=uid, index=new_doc, flush=True)
        # test value of the new document
        doc = self.zapian._get_document(uid, [part])
        values = [v.value for v in doc.values()]
        self.assertEqual(set(values), set(['946656000', 'new title']))
        # test term of the new document
        title_prefix = self.zapian.schema.get_prefix('title', auto_add=False)
        subjects_prefix = self.zapian.schema.get_prefix('subjects', auto_add=False)
        new_field_prefix = self.zapian.schema.get_prefix('new-field', auto_add=False)
        validate_terms = ['Q'+uid,
                            new_field_prefix + 'last',
                            title_prefix + 'new', 
                            title_prefix + 'title', 
                            subjects_prefix + 'file', 
                            subjects_prefix + 'ktv', 
                            subjects_prefix + 'what_gmail_com', ]
        old_terms = [term.term for term in doc.termlist()]
        self.assertTrue(len(validate_terms) == len(old_terms))
        self.assertEqual(set(validate_terms), set(old_terms))
        # test data of the new document
        self.assertEqual(doc.get_data(), '')

    def test_del_document(self):
        part = self.parts[0]
        uid = "12345"
        # add a document 
        self.zapian.add_document(part, uid=uid, 
                index=self.doc, data={'data': "测试内容"},
                flush=True)
        # delete the document
        self.zapian.delete_document(part, uids=[uid], flush=True)
        # test get the document, it will be raise KeyError
        try:
            self.zapian._get_document(uid, [part])
            raise AssertionError("Unique ID '%s' is exists" % uid)
        except KeyError:
            pass

    def test_search_document(self):
        part = self.parts[0]
        uid = "12345"
        # add a document 
        self.zapian.add_document(part, uid=uid, 
                index=self.doc, data={'data': "测试内容"},
                flush=True)
        # serach
        query = [  
                    [[u'title'], u'we', u'parse'],
                    [u'subjects', u'file ktv', u'anyof'],
                ] 
        results = self.zapian.search([part], query)
        self.assertEqual([uid], results)

    def test_search_document_for_mulit_database(self):
        """ """
        # add first document into first database
        first_uid = "12345"
        first_doc = {'title':'we are firend', 
                    'subjects':['file','ktv','what@gmail.com'] 
                    }
        self.zapian.add_document(self.parts[0], uid=first_uid, 
                index=first_doc, data={'data': "测试内容"},
                flush=True)
        # add second document into second database
        second_uid = '67890'
        second_doc = {'title':'Go to school', 
                    'subjects':['morning','walking','sport'] 
                    }
        self.zapian.add_document(self.parts[1], uid=second_uid, 
                index=second_doc, data={'data': "测试内容"},
                flush=True)
        # add third document into first database
        third_doc = {'title': 'Big Data',
                    'subjects': ['big', 'expensive']}
        self.zapian.add_document(self.parts[0], uid="45678", 
                index=third_doc, data={'data': "测试内容"},
                flush=True)

        # search for muilt database

        # search firrst document
        query = [  
                    [[u'title'], u'we', u'parse'],
                    [u'subjects', u'file ktv', u'anyof'],
                 ] 
        results = self.zapian.search(self.parts, query)
        self.assertEqual([first_uid], results)

        # search second document
        query = [  
                    [[u'title'], u'Go', u'parse'],
                    [u'subjects', u'walking sport', u'anyof'],
                ]
        results = self.zapian.search(self.parts, query)
        self.assertEqual([second_uid], results)

        # search two document
        query = [  
                    [u'title', u'Go we', u'anyof'],
                    [u'subjects', u'walking sport ktv', u'anyof'],
                ]
        results = self.zapian.search(self.parts, query)
        self.assertEqual(len(results), 2)
        self.assertEqual(set([first_uid, second_uid]), set(results))

    def test_search_document_with_exclude(self):
        self.zapian.add_document(self.parts[0], uid="12345", 
                index=self.doc, flush=True)

        new_doc = self.doc.copy()
        new_doc['title'] = 'new title'
        self.zapian.add_document(self.parts[0], uid="12346", 
                index=new_doc, flush=True)

        query = Query(self.zapian.schema)
        query.filter('subjects', u'file', 'anyof')
        results = self.zapian.search([self.parts[0]], query_obj=query)
        self.assertEqual(set(results), set(['12345', '12346']))

        query.exclude('title', u'new', 'anyof')
        results = self.zapian.search([self.parts[0]], query_obj=query)
        self.assertEqual(results, ['12345'])

    def test_multi_query(self):
        self.zapian.add_document(self.parts[0], uid="12345", 
                index=self.doc, flush=True)

        new_doc = self.doc.copy()
        new_doc['+title'] = 'we talk'
        self.zapian.add_document(self.parts[0], uid="12346", 
                index=new_doc, flush=True)

        query1 = Query(self.zapian.schema)
        query1.filter('title', u'we talk', 'anyof')
        query2 = Query(self.zapian.schema)
        query2.filter('title', u'are', '')

        combined = query2 & query1
        results = self.zapian.search([self.parts[0]], query_obj=combined)
        self.assertEqual(results, ['12345'])

        combined = query2 | query1
        results = self.zapian.search([self.parts[0]], query_obj=combined)
        self.assertEqual(set(results), set(['12345', '12346']))

if __name__ == '__main__':
    unittest.main()




# -*- coding: utf-8 -*-
"""
USAGE:

python xappy.py --<list|count|help>  < -i database > [ -n numbers, -s start ]

-l, --list      打印所有数据
可选参数: 
    -s --start   从第几条记录开始
    -n --numbers 打印多少条记录

-c, --count     列出记录数
-f, --fields    打印指定数据库中所有索引字段的名字

EXAMPLE:
    -c -i xapian.idx
    -l -i xapian.idx
    -l -i xapian.idx -s 0 -n 100
    -f -i xapian.idx
"""

import sys
import getopt
import xappy

def usage():
    print __doc__ 

def print_document_count(conn):
    """ 得到某个数据库中所有记录的个数 """
    print "共有%10d 条记录" % conn.get_doccount()

def print_documents(conn, start, limit):
    """ 打印数据库中的记录 """
    query = conn.query_all()
    if limit == 0:
        limit = conn.get_doccount()  - start
    
    results =  conn.search(query, start, limit + start)
    print "ID\tRANK\tDATA"
    print "--"*30
    for res in results:
        print '-----------'
        print "data: %s\t%s\t" % (res.id, res.rank), "; ".join([ "%s: %s" % (k, v) for k, v in res.data.items()]) 
        print 'value:    ', '; '.join([ str(value.num) + ':' + value.value for value in res._doc.values()])
        print 'term:    ', '; '.join([ value.term for value in res._doc.termlist()])

def print_fields(database):
    """ 打印数据库中已缓存的字段 """
    print "数据库中的已索引字段:"
    print "--"*15

    conn = xappy.IndexerConnection(database)
    fields = conn.get_fields_with_actions()
    for field, value in fields:
        print field, value

    print "--"*15

def main(argv):
    
    try:
        opts, args = getopt.getopt(argv, "lfci:hs:n:", ["list", "fields", "count", "index=", "help", 'start=', 'numbers='])
    except getopt.GetoptError:
        usage()
        sys.exit(2)
    
    command = None
    database = None
    list_start = 0
    list_numbers  = 0

    for opt, arg in opts:
        if opt in ('-l', '--list'):
            command = 'list'

        if opt in ('-s', '--start'):
            try:
                list_start = int(arg)
            except:
                print "start 必须是一个整数"
                sys.exit(2)

        if opt in ('-n', '--numbers'):
            try:
                list_numbers = int(arg )
            except:
                print "list_numbers 必须是一个整数"
                sys.exit(2)
    
        if opt in ('-f', '--fields'):
            command = 'fields'

        elif opt in ('-c', '--count'):
            command = 'count'

        elif opt in ('-i', '--index'):
            database = arg

        elif opt in ('-h', '--help'):
            usage()
            break;
    
    if not opts:
        usage()
        sys.exit(2)
    
    if not database:
        print "ERROR: 必须指定索引数据库的路径"

    try:
        conn = xappy.SearchConnection(database)
    except:
        print "ERROR: 无法连接指定的索引数据库"
        sys.exit(2)

    print list_start, list_numbers
    if command == 'list':
        print_documents(conn, list_start, list_numbers)

    if command == 'fields':
        print_fields(database)

    if command == 'count':
        print_document_count(conn)
            
if __name__ == '__main__':
    main(sys.argv[1:])    




# -*- encoding: utf-8 -*-

import os
import yaml

CONFIG_FILE = 'schema.yaml'

class Schema(object):
    """ The schema class is improve xapian fields
    """
    fields = None       # {'title': ’NS’ }

    attributes = None   # {'created':1 }

    def __init__(self, db_path):
        self.db_path = db_path
        self.fields = {}
        self.attributes = {}
        self.load()

    def get_prefix(self, name, auto_add=True):
        """ get prefix by field name
        """
        prefix = self.fields.get(name)
        if prefix is None and auto_add:
            prefix = self.add_field(name)
        return prefix

    def get_slot(self, name, auto_add=True):
        """ get slot by attribute name
        """
        slot = self.attributes.get(name)
        if slot is None and auto_add:
            slot = self.add_attribute(name)
        return slot

    def load(self):
        """ load fields and attributes from the yaml file """
        schema_path = os.path.join(self.db_path, CONFIG_FILE)
        if os.path.isfile(schema_path):
            with open(schema_path,'rb') as schema_file:
                schema = yaml.load(schema_file)

            self.fields = schema['fields']
            self.attributes = schema['attributes']

    def dump(self):
        """ dump fields and attributes into the yaml file """
        schema_path = os.path.join(self.db_path, CONFIG_FILE)
        schema = {'fields': self.fields, 'attributes': self.attributes}
        with open(schema_path,'wb') as schema_file:
            schema_file.write(yaml.dump(schema))

    def add_field(self, name):
        if name not in self.fields:
            prefix = self.fields[name] = self._gen_prefix()
            self.dump()
            return prefix
        else:
            return self.fields[name]

    def _gen_prefix(self):
        """ new prefix """
        count = len(self.fields) + 1
        multiple = count/24
        index = count % 24
        return 'X' + '@'*multiple + chr(index+64)

    def add_attribute(self, name):
        if name not in self.attributes:
            slot = self.attributes[name] = self._gen_slot()
            self.dump()
            return slot
        else:
            return self.attributes[name]

    def _gen_slot(self):
        """ new slot """
        return len(self.attributes) 

    def split_term(self, term):
        if term[0] == 'Q':
            return 'Q', term[1:]

        if term[0] != 'X':
            return '', term

        for index in range(1, len(term)):
            if term[index] != '@':
                break
        index += 1

        return term[:index], term[index:]






################################
import os
from setuptools import setup, find_packages

def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

setup (
    name='zapian',
    version='0.3.7',
    author = "Pan Junyong",
    author_email = "dev@zopen.cn",
    description = "indexer addons: sort, cjksplitter, and so on",
    long_description=(
        read('README')
        ),
    license = "Private",
    keywords = "zope3 z3c rpc  server client",
    classifiers = [
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Operating System :: OS Independent',
        'Topic :: Internet :: WWW/HTTP',
        'Framework :: Zope3'],
    url = 'http://github.com/everydo/zapian',
    include_package_data = True,
    namespace_packages=[],
    packages = find_packages(),
    scripts=['scripts/list_xapian_database.py'],
    install_requires = [
        'setuptools',
        'PyYAML'
        ],
    zip_safe = False,
)
